package ru.dbotthepony.mc.otm.block.entity

import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.LevelTickEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.event.level.ChunkWatchEvent
import net.minecraftforge.event.level.LevelEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import ru.dbotthepony.mc.otm.network.BlockEntitySyncPacket
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.network.WorldNetworkChannel
import ru.dbotthepony.mc.otm.onceServer
import java.lang.ref.WeakReference
import java.util.WeakHashMap

abstract class SynchronizedBlockEntity(p_155228_: BlockEntityType<*>, p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(p_155228_, p_155229_, p_155230_) {
	private var isSynchronizing = false

	val synchronizer = FieldSynchronizer {
		if (isSynchronizing)
			return@FieldSynchronizer

		if (!isRemoved && level?.isClientSide == false && (_subCache == null || (_subCache ?: throw ConcurrentModificationException()).players.isNotEmpty())) {
			onceServer {
				synchronizeToPlayers(true)
			}
		} else {
			markSynchronizerClean()
		}
	}

	private fun markSynchronizerClean() {
		synchronizer.markClean()
	}

	init {
		synchronizer.defaultEndpoint.markUnused()
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		unsubscribe()
		_subCache = null

		if (!p_155231_.isClientSide) {
			subscribe()
		}
	}

	override fun setRemoved() {
		super.setRemoved()
		unsubscribe()
	}

	override fun clearRemoved() {
		super.clearRemoved()

		if (level?.isClientSide == false) {
			subscribe()
		}
	}

	private var _subCache: ChunkSubscribers? = null

	private fun unsubscribe() {
		val subCache = _subCache ?: return

		if (subCache.unsubscribe(this)) {
			val level = subCache.level.get()

			if (level != null) {
				playerMap.get(level)?.remove(subCache.chunkPos)
			}
		}

		lastSubscriptionChangeset = -1
	}

	private fun subscribe(): ChunkSubscribers {
		val level = level
		check(level is ServerLevel) { "Invalid realm" }
		unsubscribe()

		val subs = playerMap
			.computeIfAbsent(level) { Long2ObjectAVLTreeMap() }
			.computeIfAbsent(ChunkPos(blockPos).toLong(), Long2ObjectFunction { ChunkSubscribers(level = WeakReference(level), chunkPos = ChunkPos(blockPos).toLong()) })

		subs.subscribe(this)
		_subCache = subs
		return subs
	}

	private val subscription get() = _subCache ?: subscribe()
	private var lastSubscriptionChangeset = -1

	fun synchronizeToPlayers() {
		synchronizeToPlayers(false)
	}

	protected fun synchronizeToPlayers(callbedBySynchronizer: Boolean) {
		isSynchronizing = true

		try {
			if (synchronizer.isEmpty || isRemoved) {
				if (callbedBySynchronizer) synchronizer.markClean()
				return
			}

			check(level is ServerLevel) { "Invalid realm or Level is null" }

			synchronizer.observe()

			val subscription = subscription

			if (subscription.players.isEmpty() || lastSubscriptionChangeset == subscription.changeset && !synchronizer.hasChanges) {
				if (callbedBySynchronizer) synchronizer.markClean()
				return
			}

			lastSubscriptionChangeset = subscription.changeset

			for (player in subscription.players) {
				val payload = synchronizer.computeEndpointFor(player).collectNetworkPayload()

				if (payload != null) {
					WorldNetworkChannel.send(player, BlockEntitySyncPacket(blockPos, payload.array, payload.length))
				}
			}

			synchronizer.markClean()
		} finally {
			isSynchronizing = false
		}
	}

	private data class ChunkSubscribers(
		val blockEntities: ArrayList<WeakReference<SynchronizedBlockEntity>> = ArrayList(0),
		val players: ArrayList<ServerPlayer> = ArrayList(1),
		val observingBlockEntities: ArrayList<WeakReference<SynchronizedBlockEntity>> = ArrayList(0),
		val level: WeakReference<ServerLevel>,
		val chunkPos: Long,
		var changeset: Int = 0
	) {
		fun cleanUpBlocks() {
			val listIterator = blockEntities.listIterator()

			for (block in listIterator) {
				if (block.get() == null) {
					listIterator.remove()
				}
			}
		}

		val hasObservers: Boolean get() {
			val listIterator = blockEntities.listIterator()

			for (value in listIterator) {
				val ref = value.get()

				if (ref == null) {
					listIterator.remove()
				} else if (ref.synchronizer.hasObservers) {
					return true
				}
			}

			return false
		}

		fun subscribe(blockEntity: SynchronizedBlockEntity) {
			val iterator = blockEntities.listIterator()

			for (value in iterator) {
				val ref = value.get()

				if (ref === blockEntity) {
					return
				} else if (ref == null) {
					iterator.remove()
				}
			}

			val bref = WeakReference(blockEntity)
			blockEntities.add(bref)
			changeset++

			onceServer {
				blockEntity.synchronizeToPlayers()
			}

			if (blockEntity.synchronizer.hasObservers) {
				observingBlockEntities.add(bref)
				val listing = tickingMap.computeIfAbsent(level.get() ?: throw NullPointerException("Level got GCd!")) { ArrayList(1) }
				val tickerIterator = listing.listIterator()

				for (value in tickerIterator) {
					val ref = value.get()

					if (ref === this) {
						return
					} else if (ref == null) {
						tickerIterator.remove()
					}
				}

				listing.add(WeakReference(this))
			}
		}

		fun unsubscribe(blockEntity: SynchronizedBlockEntity): Boolean {
			val listIterator = blockEntities.listIterator()

			for (value in listIterator) {
				if (value.get() === blockEntity) {
					listIterator.remove()
					changeset++
					break
				}
			}

			if (blockEntity.synchronizer.hasObservers) {
				val tickerIterator = observingBlockEntities.listIterator()

				for (value in tickerIterator) {
					val ref = value.get()

					if (ref === blockEntity) {
						tickerIterator.remove()
						break
					} else if (ref == null) {
						tickerIterator.remove()
					}
				}
			}

			if (players.isNotEmpty()) {
				return false
			}

			cleanUpBlocks()
			return blockEntities.isEmpty()
		}

		val isEmpty: Boolean get() {
			if (players.isNotEmpty()) {
				return false
			}

			cleanUpBlocks()
			return blockEntities.isEmpty()
		}
	}

	/**
	 * Why track player-tracked chunks?
	 *
	 * because minecraft itself doesn't track them, well,
	 * in "section partitioning" way.
	 *
	 * just look at [net.minecraft.server.level.PlayerMap]
	 *
	 * the [net.minecraft.server.level.PlayerMap.getPlayers] straight ignores chunk position
	 * and just fetches all players
	 *
	 * even if they did not ignore that argument, you still have to fetch player *list* though
	 * [net.minecraft.server.level.ChunkMap.getPlayers], which is not something we want
	 */
	companion object {
		private val playerMap = WeakHashMap<ServerLevel, Long2ObjectAVLTreeMap<ChunkSubscribers>>()
		private val tickingMap = WeakHashMap<ServerLevel, ArrayList<WeakReference<ChunkSubscribers>>>()

		fun onLevelUnload(event: LevelEvent.Unload) {
			val level = event.level as? ServerLevel ?: return
			playerMap.remove(level)
			tickingMap.remove(level)
		}

		fun onServerStopping(event: ServerStoppingEvent) {
			playerMap.clear()
			tickingMap.clear()
		}

		fun postLevelTick(event: LevelTickEvent) {
			if (event.phase == TickEvent.Phase.END) {
				val level = event.level as? ServerLevel ?: return
				val listing = tickingMap[level] ?: return
				val listIterator = listing.listIterator()
				var hitAnyObservers = false

				for (value in listIterator) {
					val ref = value.get()

					if (ref == null) {
						listIterator.remove()
					} else if (ref.players.isNotEmpty()) {
						var hitObservers = false
						val blockIterator = ref.observingBlockEntities.listIterator()

						for (bvalue in blockIterator) {
							val bref = bvalue.get()

							if (bref == null) {
								blockIterator.remove()
							} else {
								hitObservers = true
								hitAnyObservers = true
								bref.synchronizeToPlayers()
							}
						}

						if (!hitObservers) {
							listIterator.remove()
						}
					}
				}

				if (!hitAnyObservers) {
					tickingMap.remove(level)
				}
			}
		}

		fun onWatch(event: ChunkWatchEvent.Watch) {
			playerMap.computeIfAbsent(event.level) { Long2ObjectAVLTreeMap() }.computeIfAbsent(event.pos.toLong(), Long2ObjectFunction { ChunkSubscribers(level = WeakReference(event.level), chunkPos = event.pos.toLong()) }).let {
				val (blocks, players) = it

				if (event.player !in players) {
					players.add(event.player)
					it.changeset++

					onceServer(20) {
						if (!event.player.hasDisconnected()) {
							val iterator = blocks.listIterator()

							for (block in iterator) {
								val deref = block.get()

								if (deref == null) {
									iterator.remove()
								} else {
									deref.synchronizeToPlayers(false)
								}
							}
						}
					}
				}
			}
		}

		fun onForget(event: ChunkWatchEvent.UnWatch) {
			val subs = playerMap.get(event.level)?.get(event.pos.toLong()) ?: return

			if (subs.players.remove(event.player)) {
				if (subs.isEmpty) {
					playerMap.get(event.level)?.remove(event.pos.toLong())
				} else {
					subs.changeset++

					val iterator = subs.blockEntities.listIterator()

					for (block in iterator) {
						val deref = block.get()

						if (deref == null) {
							iterator.remove()
						} else {
							deref.synchronizer.removeEndpointFor(event.player)
						}
					}
				}
			}
		}

		fun playerDisconnected(event: PlayerEvent.PlayerLoggedOutEvent) {
			for (tree in playerMap.values) {
				val iterator = tree.iterator()

				for (entry in iterator) {
					val listIterator = entry.value.players.listIterator()

					for (value in listIterator) {
						if (value === event.entity) {
							listIterator.remove()
							entry.value.changeset++
						}
					}

					if (entry.value.isEmpty) {
						iterator.remove()
					}
				}
			}
		}
	}
}
