package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.core.set

enum class SkinElementType {
	SINGLE {
		override fun toJson(value: AbstractSkinElement): JsonObject {
			require(value is SkinElement) { "Invalid skin element provided, expected SkinElement, got ${value::class.qualifiedName}" }

			return JsonObject().also {
				it["texture"] = JsonPrimitive(value.texture.toString())
				it["x"] = JsonPrimitive(value.x)
				it["y"] = JsonPrimitive(value.y)
				it["width"] = JsonPrimitive(value.width)
				it["height"] = JsonPrimitive(value.height)
				it["imageWidth"] = JsonPrimitive(value.imageWidth)
				it["imageHeight"] = JsonPrimitive(value.imageHeight)
				it["winding"] = JsonPrimitive(value.winding.name)
			}
		}

		override fun toNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf) {
			require(value is SkinElement) { "Invalid skin element provided, expected SkinElement, got ${value::class.qualifiedName}" }

			buff.writeUtf(value.texture.toString())
			buff.writeFloat(value.x)
			buff.writeFloat(value.y)
			buff.writeFloat(value.width)
			buff.writeFloat(value.height)
			buff.writeFloat(value.imageWidth)
			buff.writeFloat(value.imageHeight)
			buff.writeEnum(value.winding)
		}

		override fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement {
			val texture = ResourceLocation(buff.readUtf())
			val x = buff.readFloat()
			val y = buff.readFloat()
			val width = buff.readFloat()
			val height = buff.readFloat()
			val imageWidth = buff.readFloat()
			val imageHeight = buff.readFloat()
			val winding = buff.readEnum(UVWindingOrder::class.java)

			return SkinElement(texture, x, y, width, height, imageWidth, imageHeight, winding)
		}

		override fun fromJson(element: JsonObject): AbstractSkinElement {
			val texture = element["texture"]?.asString ?: throw JsonSyntaxException("Missing texture element")
			val x = element["x"]?.asFloat ?: throw JsonSyntaxException("Missing x element")
			val y = element["y"]?.asFloat ?: throw JsonSyntaxException("Missing y element")
			val width = element["width"]?.asFloat ?: throw JsonSyntaxException("Missing width element")
			val height = element["height"]?.asFloat ?: throw JsonSyntaxException("Missing height element")
			val imageWidth = element["imageWidth"]?.asFloat ?: throw JsonSyntaxException("Missing imageWidth element")
			val imageHeight = element["imageHeight"]?.asFloat ?: throw JsonSyntaxException("Missing imageHeight element")
			val winding = element["winding"]?.asString ?: throw JsonSyntaxException("Missing winding element")

			return SkinElement(ResourceLocation.tryParse(texture) ?: throw JsonSyntaxException("Invalid resource location: $texture"), x, y, width, height, imageWidth, imageHeight, UVWindingOrder.valueOf(winding))
		}
	},
	ATLAS {
		override fun toJson(value: AbstractSkinElement): JsonObject {
			require(value is AtlasSkinElement) { "Invalid skin element provided, expected AtlasSkinElement, got ${value::class.qualifiedName}" }

			return JsonObject().also {
				it["location"] = JsonPrimitive(value.location.toString())
				it["width"] = JsonPrimitive(value.width)
				it["height"] = JsonPrimitive(value.height)
			}
		}

		override fun toNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf) {
			require(value is AtlasSkinElement) { "Invalid skin element provided, expected AtlasSkinElement, got ${value::class.qualifiedName}" }

			buff.writeResourceLocation(value.location)
			buff.writeFloat(value.width)
			buff.writeFloat(value.height)
		}

		override fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement {
			val location = ResourceLocation(buff.readUtf())
			val width = buff.readFloat()
			val height = buff.readFloat()

			return AtlasSkinElement(location, width, height)
		}

		override fun fromJson(element: JsonObject): AbstractSkinElement {
			val location = element["location"]?.asString ?: throw JsonSyntaxException("Missing location element")
			val width = element["width"]?.asFloat ?: throw JsonSyntaxException("Missing width element")
			val height = element["height"]?.asFloat ?: throw JsonSyntaxException("Missing height element")

			return AtlasSkinElement(ResourceLocation.tryParse(location) ?: throw JsonSyntaxException("Invalid resource location: $location"), width, height)
		}
	},
	SUBELEMENT {
		override fun toJson(value: AbstractSkinElement): JsonObject {
			require(value is SubSkinElement) { "Invalid skin element provided, expected SubSkinElement, got ${value::class.qualifiedName}" }

			return JsonObject().also {
				it["parent"] = value.parent.toJson()
				it["xOffset"] = JsonPrimitive(value.xOffset)
				it["yOffset"] = JsonPrimitive(value.yOffset)
				it["subWidth"] = JsonPrimitive(value.width)
				it["subHeight"] = JsonPrimitive(value.height)
			}
		}

		override fun toNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf) {
			require(value is SubSkinElement) { "Invalid skin element provided, expected SubSkinElement, got ${value::class.qualifiedName}" }

			value.parent.toNetwork(buff)
			buff.writeFloat(value.xOffset)
			buff.writeFloat(value.yOffset)
			buff.writeFloat(value.width)
			buff.writeFloat(value.height)
		}

		override fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement {
			val parent = Companion.fromNetwork(buff)
			val xOffset = buff.readFloat()
			val yOffset = buff.readFloat()
			val subWidth = buff.readFloat()
			val subHeight = buff.readFloat()

			return SubSkinElement(parent, xOffset, yOffset, subWidth, subHeight)
		}

		override fun fromJson(element: JsonObject): AbstractSkinElement {
			val parent = element["parent"] as? JsonObject ?: throw JsonSyntaxException("Invalid parent")
			val xOffset = element["xOffset"].asFloat
			val yOffset = element["yOffset"].asFloat
			val subWidth = element["subWidth"].asFloat
			val subHeight = element["subHeight"].asFloat

			return SubSkinElement(Companion.fromJson(parent), xOffset, yOffset, subWidth, subHeight)
		}
	};

	protected abstract fun toJson(value: AbstractSkinElement): JsonObject
	protected abstract fun toNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf)
	protected abstract fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement
	protected abstract fun fromJson(element: JsonObject): AbstractSkinElement

	fun toActualJson(value: AbstractSkinElement): JsonObject {
		return toJson(value).also {
			it["type"] = JsonPrimitive(name)
		}
	}

	fun toActualNetwork(value: AbstractSkinElement, buff: FriendlyByteBuf) {
		buff.writeEnum(this)
		toNetwork(value, buff)
	}

	companion object {
		fun fromNetwork(buff: FriendlyByteBuf): AbstractSkinElement {
			val type = buff.readEnum(SkinElementType::class.java)
			return type.fromNetwork(buff)
		}

		fun fromJson(element: JsonObject): AbstractSkinElement {
			val type = SkinElementType.valueOf(element["type"].asString)
			return type.fromJson(element)
		}
	}
}
